/*!
 * backbone.localcache.js
 */

define(function (require) {


/* -----------------------------------------------------------------------------
 * dependencies
 * ---------------------------------------------------------------------------*/

var _ = require('underscore');
var mixin = require('mixins/mixin');
var Cache = require('cache/cache');


/* -----------------------------------------------------------------------------
 * backboneLocalcache
 * ---------------------------------------------------------------------------*/

return {

  attachments: ['meta'],

  _flags: {
    restoredFromCache: false
  },

  constructor: mixin('decorate', function (fn, attrs, options) {
    options = options || {};

    // cached collections will pass model attributes through options
    this.__uuid = this._getUUID(options);
    this.cacheName = options.cacheName || this.cacheName;
    this.cacheOptions = _.extend({}, this.cacheOptions, options.cacheOptions);
    this.bucketName = options.bucketName || this.bucketName;

    // before
    // ==================================================
    if (!this.cacheName) {
      throw new Error('Required attribute `cacheName` missing.');
    } else {
      this._createCache(attrs, options);
    }

    // retrieve attrs (specifically uuid) prior to adding attachments
    attrs = this._mergeAttrs(this._loadCacheAttrs(), attrs);
    this.__uuid = this._isModel(this) ? this._getUUID(attrs, true) : null;
    this.__attachments = this._loadAttachments(options);

    // constructor
    // ==================================================
    fn.call(this, attrs, options);

    // after
    // ==================================================
    this._syncCacheAttrs();

    // delete temporarily stored cached attachments as the should have been
    // used by this point
    delete this.__attachments;

    this._syncAttachments(this.attachments);
    this._addAttachmentListeners(this.attachments);
  }),

  _prepareModel: mixin('decorate', function (fn, attrs, options) {
    var cloned = _.pick(this, 'cacheName', 'cacheOptions', 'bucketName');
    var uuid = this._getUUID(attrs, true);
    var model = fn.call(this, attrs, _.extend(cloned, { __uuid: uuid }, options));

    // if created model wasn't a cacheModel it ignored the uuid we just passed
    // to its constructor. As a result we will attach it here to be safe
    model.__uuid = uuid;
    return model;
  }),

  restoredFromCache: function () {
    return this._flags.restoredFromCache;
  },


  /* ---------------------------------------------------------------------------
   * configurable methods
   * -------------------------------------------------------------------------*/

  _getCacheOptions: function (attrs, options) {
    var defaults = this._getCacheDefaults(attrs, options);
    return _.extend(defaults, this.cacheOptions || {});
  },

  _getCacheDefaults: function (attrs, options) {
    this.cacheDefaults = this.cacheDefaults || this.getCacheDefaults(attrs, options);
    return this.cacheDefaults;
  },

  // overwrite to alter defaults creation logic
  getCacheDefaults: function (attrs, options) {
    return { bucket: this._getBucketName(attrs, options) };
  },

  _getBucketName: function (attrs, options) {
    this.bucketName = this.bucketName || this.getBucketName(attrs, options);
    return this.bucketName;
  },

  // overwrite to alter bucketName creation logic
  getBucketName: function (attrs, options) {
    return this.cacheName;
  },


  /* ---------------------------------------------------------------------------
   * event handlers
   * -------------------------------------------------------------------------*/

  _onAddToCollection: function (model, collection, options) {
    if (!options || !options.ignoreCache) {
      this._setCacheEntry(model);
    }
  },

  _onRemoveFromCollection: function (model, collection, options) {
    if (!options || !options.ignoreCache) {
      this.cache.remove(model.id);
    }
  },

  _onResetCollection: function (collection, options) {
    if (!options || !options.ignoreCache) {
      this.cache.flush();
    }
  },

  _onChangeModel: function (model, options) {
    if (!options || !options.ignoreCache) {
      this._setCacheEntry(model);
    }
  },


  /* ---------------------------------------------------------------------------
   * cache
   * -------------------------------------------------------------------------*/

  _createCache: function (attrs, options) {
    this.cache = new Cache(this._getCacheOptions(attrs, options));
    this.cache.flushExpired();

    this.on('add', this._onAddToCollection, this);
    this.on('remove', this._onRemoveFromCollection, this);
    this.on('reset', this._onResetCollection, this);
    this.on('change', this._onChangeModel, this);
  },

  _loadCacheAttrs: function () {
    var cacheAttrs = this._getCacheAttrs();

    this._flags.restoredFromCache = !_.isEmpty(cacheAttrs);

    return this._isModel(this) && _.isArray(cacheAttrs)
      ? cacheAttrs[0]
      : cacheAttrs;
  },

  _getCacheAttrs: function () {
    return this.__uuid
      ? this._getFromCacheByUUID()
      : this._getFromCacheByFilter();
  },

  _getFromCacheByUUID: function () {
    return this.cache.get(this._getCacheKey(this));
  },

  _getFromCacheByFilter: function () {
    // lets cache attachment keys right before filtering
    this.attachmentKeys = _.object(this.attachments, []);

    return _.chain(this.cache.all())
      .pick(this._filterCacheAttrs.bind(this))
      .map(this._attachUUID.bind(this))
      .value();
  },

  _filterCacheAttrs: function (val, key) {
    var attachmentIndex = key.lastIndexOf('__');

    // do I even need to check it against attachmentKeys? or is the pure existance
    // of the attachment tell me enough?
    return attachmentIndex !== -1
      ? !this.attachmentKeys.hasOwnProperty(key.slice(attachmentIndex + 2))
      : true;
  },

  _attachUUID: function (val, key) {
    var attrs = JSON.parse(val);

    attrs.__uuid = key;
    return attrs;
  },

  _syncCacheAttrs: function () {
    return this.models
      ? _.each(this.models, this._setCacheEntry, this)
      : this._setCacheEntry(this);
  },

  _setCacheEntry: function (model) {
    var cacheKey = this._getCacheKey(model);

    if (cacheKey) {
      return _.isEmpty(model.attributes)
        ? this.cache.remove(cacheKey)
        : this.cache.set(cacheKey, model.attributes);
    }
  },

  _mergeAttrs: function (cacheAttrs, attrs) {
    if (!attrs) {
      return cacheAttrs;
    } else if (_.isEmpty(cacheAttrs)) {
      return attrs;
    }

    return this._isModel(this)
      ? this._mergeModelAttrs(cacheAttrs, attrs)
      : this._mergeCollectionAttrs(cacheAttrs, attrs);
  },

  _mergeModelAttrs: function (cacheAttrs, attrs) {
    return _.extend(cacheAttrs, attrs);
  },

  _mergeCollectionAttrs: function (cacheAttrs, attrs) {
    var uniqItaratee = function (val) {
      return val[this.modelId];
    }.bind(this);

    return (_.isArray(attrs) && _.isArray(cacheAttrs))
      ? _.uniq(_.union(attrs, cacheAttrs), uniqItaratee)
      : _.uniq([attrs].concat(cacheAttrs), uniqItaratee);
  },


  /* ---------------------------------------------------------------------------
   * attachments
   * -------------------------------------------------------------------------*/

  _loadAttachments: function (options) {
    var attachments = _.reduce(this.attachments, function (attachments, name) {
      attachments[name] = _.extend({}, this._loadAttachment(name), options[name] || {});
      return attachments;
    }, {}, this);

    return attachments;
  },

  _loadAttachment: function (name) {
    var cacheKey = this._getCacheKey(this, name);
    return this.cache.get(cacheKey);
  },

  _syncAttachments: function (attachments) {
    _.each(attachments, function (name) {
      this._syncAttachment(name, this[name]);
    }, this);
  },

  _syncAttachment: function (name, attachment) {
    if (!attachment) { return; }

    var cacheKey = this._getCacheKey(this, name);
    return this.cache.set(cacheKey, attachment.attributes);
  },

  _addAttachmentListeners: function (attachments) {
    _.each(attachments, function (name) {
      this.on(name + ':change', this._onChangeAttachment.bind(this, name));
    }, this);
  },

  _onChangeAttachment: function (name, attachment, options) {
    if ((!options || !options.ignoreCache) && this[name] === attachment) {
      this._syncAttachment(name, attachment);
    }
  },


  /* ---------------------------------------------------------------------------
   * helpers
   * -------------------------------------------------------------------------*/

  _isModel: function (entity) {
    return entity instanceof Backbone.Model;
  },

  _getUUID: function (obj, shouldGenerate) {
    obj = obj || {};

    var uuid = obj.__uuid || this.__uuid;
    delete obj.__uuid;

    return !uuid && shouldGenerate
      ? this._generateUUID()
      : uuid;
  },

  _generateUUID: function () {
    var d = new Date().getTime();
    var uuid = 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
      var r = (d + Math.random()*16)%16 | 0;
      d = Math.floor(d/16);
      return (c === 'x' ? r : (r&0x3|0x8)).toString(16);
    });

    return uuid;
  },

  _getCacheKey: function (entity, name) {
    if (!this._isModel(entity)) {
      return '__' + name;
    }

    return name
      ? entity.__uuid + '-__' + name
      : entity.__uuid;
  }

};

});
