(function (root, factory) {
  if (typeof define === 'function' && define.amd) {
    // AMD. Register as an anonymous module.
    define(['underscore'], function (underscore) {
      return (root.returnExportsGlobal = factory(underscore));
    });
  } else if (typeof exports === 'object') {
    // Node. Does not work with strict CommonJS, but
    // only CommonJS-like enviroments that support module.exports,
    // like Node.
    module.exports = factory(require('underscore'));
  } else {
    root['backboneLocalcache'] = factory(root.underscore);
  }
}(this, function (underscore) {

/*!
 * mixin.js
 */
var mixinsMixin, cacheCache, backbonelocalcache, _lscache_;
mixinsMixin = function (require) {
  /* -----------------------------------------------------------------------------
   * mixin
   * helper fn used to create mixins.js compatible mixins.
   * ---------------------------------------------------------------------------*/
  /**
   * @public
   * @memberof mixins
   *
   * @desc Method to attach mixin meta data to objects. This meta data is used
   * by the mixer in order to determine how to merge the given object.
   *
   * @example
   * var myMixin = {
   *   ui: mixins.mixin('defaults', { $btn: '.default-btn' })
   * };
   *
   * @param {string} type - The mixin type to attach to the object
   * @param {Object|Array|Function} obj - The object which will be merged.
   */
  var mixin = function (type, obj) {
    Object.defineProperty(obj, '__mixin', {
      enumerable: false,
      writable: true
    });
    obj.__mixin = type;
    return obj;
  };
  /* -----------------------------------------------------------------------------
   * expose
   * ---------------------------------------------------------------------------*/
  return mixin;
}({});
/**
 * lscache library
 * Copyright (c) 2011, Pamela Fox
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/* jshint undef:true, browser:true, node:true */
/* global define */
(function (root, factory) {
  if (true) {
    // AMD. Register as an anonymous module.
    _lscache_ = function () {
      return typeof factory === 'function' ? factory() : factory;
    }();
  } else if (typeof module !== 'undefined' && module.exports) {
    // CommonJS/Node module
    module.exports = factory();
  } else {
    // Browser globals
    root.lscache = factory();
  }
}(this, function () {
  // scope
  var cachedStorage;
  var cachedJSON;
  var cacheBucket = '';
  var warnings = false;
  var lscache = {
    // Prefix for all lscache keys
    PREFIX: 'lscache-',
    // Suffix for the key name on the expiration items in localStorage
    SUFFIX: '-cacheexpiration',
    // expiration date radix (set to Base-36 for most space savings)
    EXPIRY_RADIX: 10,
    // time resolution in minutes
    EXPIRY_UNITS: 60 * 1000,
    // Determines if localStorage is supported in the browser;
    // result is cached for better performance instead of being run each time.
    // Feature detection is based on how Modernizr does it;
    // it's not straightforward due to FF4 issues.
    // It's not run at parse-time as it takes 200ms in Android.
    _supportsStorage: function () {
      var key = '__lscachetest__';
      var value = key;
      if (cachedStorage !== undefined) {
        return cachedStorage;
      }
      try {
        lscache._setItem(key, value);
        lscache._removeItem(key);
        cachedStorage = true;
      } catch (e) {
        if (lscache._isOutOfSpace(e)) {
          // If we hit the limit, then it means we have support, 
          cachedStorage = true;  // just maxed it out and even the set test failed.
        } else {
          cachedStorage = false;
        }
      }
      return cachedStorage;
    },
    // Check to set if the error is us dealing with being out of space
    _isOutOfSpace: function (e) {
      if (e && e.name === 'QUOTA_EXCEEDED_ERR' || e.name === 'NS_ERROR_DOM_QUOTA_REACHED' || e.name === 'QuotaExceededError') {
        return true;
      }
      return false;
    },
    // Determines if native JSON (de-)serialization is supported in the browser.
    _supportsJSON: function () {
      /*jshint eqnull:true */
      if (cachedJSON === undefined) {
        cachedJSON = window.JSON != null;
      }
      return cachedJSON;
    },
    /**
     * Returns a string where all RegExp special characters are escaped with a \.
     * @param {String} text
     * @return {string}
     */
    _escapeRegExpSpecialCharacters: function (text) {
      return text.replace(/[[\]{}()*+?.\\^$|]/g, '\\$&');
    },
    /**
     * Returns the full string for the localStorage expiration item.
     * @param {String} key
     * @return {string}
     */
    _expirationKey: function (key) {
      return key + lscache.SUFFIX;
    },
    /**
     * Returns the number of minutes since the epoch.
     * @return {number}
     */
    _currentTime: function () {
      return Math.floor(new Date().getTime() / lscache.EXPIRY_UNITS);
    },
    /**
     * ECMAScript max Date (epoch + 1e8 days)
     */
    _getMaxDate: function () {
      lscache._maxDate = lscache._maxDate || Math.floor(8640000000000000 / lscache.EXPIRY_UNITS);
      return lscache._maxDate;
    },
    /**
     * Wrapper functions for localStorage methods
     */
    _getItem: function (key) {
      return localStorage.getItem(lscache.PREFIX + cacheBucket + key);
    },
    _setItem: function (key, value) {
      // Fix for iPad issue - sometimes throws QUOTA_EXCEEDED_ERR on setItem.
      localStorage.removeItem(lscache.PREFIX + cacheBucket + key);
      localStorage.setItem(lscache.PREFIX + cacheBucket + key, value);
    },
    _removeItem: function (key) {
      localStorage.removeItem(lscache.PREFIX + cacheBucket + key);
    },
    _eachKey: function (fn) {
      var prefixRegExp = new RegExp('^' + lscache.PREFIX + lscache._escapeRegExpSpecialCharacters(cacheBucket) + '(.*)');
      // Loop in reverse as removing items will change indices of tail
      for (var i = localStorage.length - 1; i >= 0; --i) {
        var key = localStorage.key(i);
        key = key && key.match(prefixRegExp);
        key = key && key[1];
        if (key && key.indexOf(lscache.SUFFIX) < 0) {
          fn(key, lscache._expirationKey(key));
        }
      }
    },
    _flushItem: function (key) {
      var exprKey = lscache._expirationKey(key);
      lscache._removeItem(key);
      lscache._removeItem(exprKey);
    },
    _flushExpiredItem: function (key) {
      var exprKey = lscache._expirationKey(key);
      var expr = lscache._getItem(exprKey);
      if (expr) {
        var expirationTime = parseInt(expr, lscache.EXPIRY_RADIX);
        // Check if we should actually kick item out of storage
        if (lscache._currentTime() >= expirationTime) {
          lscache._removeItem(key);
          lscache._removeItem(exprKey);
          return true;
        }
      }
    },
    _warn: function (message, err) {
      if (!warnings)
        return;
      if (!('console' in window) || typeof window.console.warn !== 'function')
        return;
      window.console.warn('lscache - ' + message);
      if (err)
        window.console.warn('lscache - The error was: ' + err.message);
    },
    /**
     * Stores the value in localStorage. Expires after specified number of minutes.
     * @param {string} key
     * @param {Object|string} value
     * @param {number} time
     */
    set: function (key, value, time) {
      if (!lscache._supportsStorage())
        return;
      // If we don't get a string value, try to stringify
      // In future, localStorage may properly support storing non-strings
      // and this can be removed.
      if (typeof value !== 'string') {
        if (!lscache._supportsJSON())
          return;
        try {
          value = JSON.stringify(value);
        } catch (e) {
          // Sometimes we can't stringify due to circular refs
          // in complex objects, so we won't bother storing then.
          return;
        }
      }
      try {
        lscache._setItem(key, value);
      } catch (e) {
        if (lscache._isOutOfSpace(e)) {
          // If we exceeded the quota, then we will sort
          // by the expire time, and then remove the N oldest
          var storedKeys = [];
          var storedKey;
          lscache._eachKey(function (key, exprKey) {
            var expiration = lscache._getItem(exprKey);
            if (expiration) {
              expiration = parseInt(expiration, lscache.EXPIRY_RADIX);
            } else {
              // TODO: Store date added for non-expiring items for smarter removal
              expiration = lscache._getMaxDate();
            }
            storedKeys.push({
              key: key,
              size: (lscache._getItem(key) || '').length,
              expiration: expiration
            });
          });
          // Sorts the keys with oldest expiration time last
          storedKeys.sort(function (a, b) {
            return b.expiration - a.expiration;
          });
          var targetSize = (value || '').length;
          while (storedKeys.length && targetSize > 0) {
            storedKey = storedKeys.pop();
            lscache._warn('Cache is full, removing item with key \'' + key + '\'');
            lscache._flushItem(storedKey.key);
            targetSize -= storedKey.size;
          }
          try {
            lscache._setItem(key, value);
          } catch (e) {
            // value may be larger than total quota
            lscache._warn('Could not add item with key \'' + key + '\', perhaps it\'s too big?', e);
            return;
          }
        } else {
          // If it was some other error, just give up.
          lscache._warn('Could not add item with key \'' + key + '\'', e);
          return;
        }
      }
      // If a time is specified, store expiration info in localStorage
      if (time) {
        lscache._setItem(lscache._expirationKey(key), (lscache._currentTime() + time).toString(lscache.EXPIRY_RADIX));
      } else {
        // In case they previously set a time, remove that info from localStorage.
        lscache._removeItem(lscache._expirationKey(key));
      }
    },
    /**
     * Retrieves specified value from localStorage, if not expired.
     * @param {string} key
     * @return {string|Object}
     */
    get: function (key) {
      if (!lscache._supportsStorage())
        return null;
      // Return the de-serialized item if not expired
      if (lscache._flushExpiredItem(key)) {
        return null;
      }
      // Tries to de-serialize stored value if its an object, and returns the normal value otherwise.
      var value = lscache._getItem(key);
      if (!value || !lscache._supportsJSON()) {
        return value;
      }
      try {
        // We can't tell if its JSON or a string, so we try to parse
        return JSON.parse(value);
      } catch (e) {
        // If we can't parse, it's probably because it isn't an object
        return value;
      }
    },
    /**
     * Removes a value from localStorage.
     * Equivalent to 'delete' in memcache, but that's a keyword in JS.
     * @param {string} key
     */
    remove: function (key) {
      if (!lscache._supportsStorage())
        return;
      lscache._flushItem(key);
    },
    /**
     * Returns whether local storage is supported.
     * Currently exposed for testing purposes.
     * @return {boolean}
     */
    supported: function () {
      return lscache._supportsStorage();
    },
    /**
     * Flushes all lscache items and expiry markers without affecting rest of localStorage
     */
    flush: function () {
      if (!lscache._supportsStorage())
        return;
      lscache._eachKey(function (key) {
        lscache._flushItem(key);
      });
    },
    /**
     * Flushes expired lscache items and expiry markers without affecting rest of localStorage
     */
    flushExpired: function () {
      if (!lscache._supportsStorage())
        return;
      lscache._eachKey(function (key) {
        lscache._flushExpiredItem(key);
      });
    },
    /**
     * Appends lscache.PREFIX so lscache will partition data in to different buckets.
     * @param {string} bucket
     */
    setBucket: function (bucket) {
      cacheBucket = bucket;
    },
    /**
     * Resets the string being appended to lscache.PREFIX so lscache will use the default storage behavior.
     */
    resetBucket: function () {
      cacheBucket = '';
    },
    /**
     * Sets whether to display warnings when an item is removed from the cache or not.
     */
    enableWarnings: function (enabled) {
      warnings = enabled;
    }
  };
  // Return the module
  return lscache;
}));
/*!
 * cache.js
 */
cacheCache = function (require) {
  /* -----------------------------------------------------------------------------
   * dependencies
   * ---------------------------------------------------------------------------*/
  var lscache = _lscache_;
  /* -----------------------------------------------------------------------------
   * Cache
   *
   * Cache class with ttl, namespace, and bucket support (wrapper around lscache).
   * ---------------------------------------------------------------------------*/
  var Cache = function (options) {
    options = options || {};
    this.namespace = options.namespace;
    this.bucket = options.bucket;
    this.ttl = options.ttl;
  };
  Cache.prototype.bucketId = function () {
    var id = '';
    if (this.namespace) {
      id += this.namespace + '-';
    }
    if (this.bucket) {
      id += this.bucket + '-';
    }
    return id;
  };
  Cache.prototype.get = function () {
    lscache.setBucket(this.bucketId());
    return lscache.get.apply(lscache, arguments);
  };
  Cache.prototype.set = function (key, val, ttl) {
    lscache.setBucket(this.bucketId());
    return lscache.set.call(lscache, key, val, ttl || this.ttl);
  };
  Cache.prototype.remove = function () {
    lscache.setBucket(this.bucketId());
    return lscache.remove.apply(lscache, arguments);
  };
  Cache.prototype.flush = function () {
    lscache.setBucket(this.bucketId());
    return lscache.flush.apply(lscache, arguments);
  };
  Cache.prototype.flushExpired = function () {
    lscache.setBucket(this.bucketId());
    return lscache.flushExpired.apply(lscache, arguments);
  };
  Cache.prototype.all = function () {
    var data = {};
    lscache.setBucket(this.bucketId());
    lscache._eachKey(function (key, exprKey) {
      data[key] = lscache._getItem(key);
    });
    return data;
  };
  /* -----------------------------------------------------------------------------
   * expose
   * ---------------------------------------------------------------------------*/
  return Cache;
}({});
/*!
 * backbone.localcache.js
 */
backbonelocalcache = function (require) {
  /* -----------------------------------------------------------------------------
   * dependencies
   * ---------------------------------------------------------------------------*/
  var _ = underscore;
  var mixin = mixinsMixin;
  var Cache = cacheCache;
  /* -----------------------------------------------------------------------------
   * backboneLocalcache
   * ---------------------------------------------------------------------------*/
  return {
    attachments: ['meta'],
    _flags: { restoredFromCache: false },
    constructor: mixin('decorate', function (fn, attrs, options) {
      options = options || {};
      // cached collections will pass model attributes through options
      this.__uuid = this._getUUID(options);
      this.cacheName = options.cacheName || this.cacheName;
      this.cacheOptions = _.extend({}, this.cacheOptions, options.cacheOptions);
      this.bucketName = options.bucketName || this.bucketName;
      // before
      // ==================================================
      if (!this.cacheName) {
        throw new Error('Required attribute `cacheName` missing.');
      } else {
        this._createCache(attrs, options);
      }
      // retrieve attrs (specifically uuid) prior to adding attachments
      attrs = this._mergeAttrs(this._loadCacheAttrs(), attrs);
      this.__uuid = this._isModel(this) ? this._getUUID(attrs, true) : null;
      this.__attachments = this._loadAttachments(options);
      // constructor
      // ==================================================
      fn.call(this, attrs, options);
      // after
      // ==================================================
      this._syncCacheAttrs();
      // delete temporarily stored cached attachments as the should have been
      // used by this point
      delete this.__attachments;
      this._syncAttachments(this.attachments);
      this._addAttachmentListeners(this.attachments);
    }),
    _prepareModel: mixin('decorate', function (fn, attrs, options) {
      var cloned = _.pick(this, 'cacheName', 'cacheOptions', 'bucketName');
      var uuid = this._getUUID(attrs, true);
      var model = fn.call(this, attrs, _.extend(cloned, { __uuid: uuid }, options));
      // if created model wasn't a cacheModel it ignored the uuid we just passed
      // to its constructor. As a result we will attach it here to be safe
      model.__uuid = uuid;
      return model;
    }),
    restoredFromCache: function () {
      return this._flags.restoredFromCache;
    },
    /* ---------------------------------------------------------------------------
     * configurable methods
     * -------------------------------------------------------------------------*/
    _getCacheOptions: function (attrs, options) {
      var defaults = this._getCacheDefaults(attrs, options);
      return _.extend(defaults, this.cacheOptions || {});
    },
    _getCacheDefaults: function (attrs, options) {
      this.cacheDefaults = this.cacheDefaults || this.getCacheDefaults(attrs, options);
      return this.cacheDefaults;
    },
    // overwrite to alter defaults creation logic
    getCacheDefaults: function (attrs, options) {
      return { bucket: this._getBucketName(attrs, options) };
    },
    _getBucketName: function (attrs, options) {
      this.bucketName = this.bucketName || this.getBucketName(attrs, options);
      return this.bucketName;
    },
    // overwrite to alter bucketName creation logic
    getBucketName: function (attrs, options) {
      return this.cacheName;
    },
    /* ---------------------------------------------------------------------------
     * event handlers
     * -------------------------------------------------------------------------*/
    _onAddToCollection: function (model, collection, options) {
      if (!options || !options.ignoreCache) {
        this._setCacheEntry(model);
      }
    },
    _onRemoveFromCollection: function (model, collection, options) {
      if (!options || !options.ignoreCache) {
        this.cache.remove(model.id);
      }
    },
    _onResetCollection: function (collection, options) {
      if (!options || !options.ignoreCache) {
        this.cache.flush();
      }
    },
    _onChangeModel: function (model, options) {
      if (!options || !options.ignoreCache) {
        this._setCacheEntry(model);
      }
    },
    /* ---------------------------------------------------------------------------
     * cache
     * -------------------------------------------------------------------------*/
    _createCache: function (attrs, options) {
      this.cache = new Cache(this._getCacheOptions(attrs, options));
      this.cache.flushExpired();
      this.on('add', this._onAddToCollection, this);
      this.on('remove', this._onRemoveFromCollection, this);
      this.on('reset', this._onResetCollection, this);
      this.on('change', this._onChangeModel, this);
    },
    _loadCacheAttrs: function () {
      var cacheAttrs = this._getCacheAttrs();
      this._flags.restoredFromCache = !_.isEmpty(cacheAttrs);
      return this._isModel(this) && _.isArray(cacheAttrs) ? cacheAttrs[0] : cacheAttrs;
    },
    _getCacheAttrs: function () {
      return this.__uuid ? this._getFromCacheByUUID() : this._getFromCacheByFilter();
    },
    _getFromCacheByUUID: function () {
      return this.cache.get(this._getCacheKey(this));
    },
    _getFromCacheByFilter: function () {
      // lets cache attachment keys right before filtering
      this.attachmentKeys = _.object(this.attachments, []);
      return _.chain(this.cache.all()).pick(this._filterCacheAttrs.bind(this)).map(this._attachUUID.bind(this)).value();
    },
    _filterCacheAttrs: function (val, key) {
      var attachmentIndex = key.lastIndexOf('__');
      // do I even need to check it against attachmentKeys? or is the pure existance
      // of the attachment tell me enough?
      return attachmentIndex !== -1 ? !this.attachmentKeys.hasOwnProperty(key.slice(attachmentIndex + 2)) : true;
    },
    _attachUUID: function (val, key) {
      var attrs = JSON.parse(val);
      attrs.__uuid = key;
      return attrs;
    },
    _syncCacheAttrs: function () {
      return this.models ? _.each(this.models, this._setCacheEntry, this) : this._setCacheEntry(this);
    },
    _setCacheEntry: function (model) {
      var cacheKey = this._getCacheKey(model);
      if (cacheKey) {
        return _.isEmpty(model.attributes) ? this.cache.remove(cacheKey) : this.cache.set(cacheKey, model.attributes);
      }
    },
    _mergeAttrs: function (cacheAttrs, attrs) {
      if (!attrs) {
        return cacheAttrs;
      } else if (_.isEmpty(cacheAttrs)) {
        return attrs;
      }
      return this._isModel(this) ? this._mergeModelAttrs(cacheAttrs, attrs) : this._mergeCollectionAttrs(cacheAttrs, attrs);
    },
    _mergeModelAttrs: function (cacheAttrs, attrs) {
      return _.extend(cacheAttrs, attrs);
    },
    _mergeCollectionAttrs: function (cacheAttrs, attrs) {
      var uniqItaratee = function (val) {
        return val[this.modelId];
      }.bind(this);
      return _.isArray(attrs) && _.isArray(cacheAttrs) ? _.uniq(_.union(attrs, cacheAttrs), uniqItaratee) : _.uniq([attrs].concat(cacheAttrs), uniqItaratee);
    },
    /* ---------------------------------------------------------------------------
     * attachments
     * -------------------------------------------------------------------------*/
    _loadAttachments: function (options) {
      var attachments = _.reduce(this.attachments, function (attachments, name) {
        attachments[name] = _.extend({}, this._loadAttachment(name), options[name] || {});
        return attachments;
      }, {}, this);
      return attachments;
    },
    _loadAttachment: function (name) {
      var cacheKey = this._getCacheKey(this, name);
      return this.cache.get(cacheKey);
    },
    _syncAttachments: function (attachments) {
      _.each(attachments, function (name) {
        this._syncAttachment(name, this[name]);
      }, this);
    },
    _syncAttachment: function (name, attachment) {
      if (!attachment) {
        return;
      }
      var cacheKey = this._getCacheKey(this, name);
      return this.cache.set(cacheKey, attachment.attributes);
    },
    _addAttachmentListeners: function (attachments) {
      _.each(attachments, function (name) {
        this.on(name + ':change', this._onChangeAttachment.bind(this, name));
      }, this);
    },
    _onChangeAttachment: function (name, attachment, options) {
      if ((!options || !options.ignoreCache) && this[name] === attachment) {
        this._syncAttachment(name, attachment);
      }
    },
    /* ---------------------------------------------------------------------------
     * helpers
     * -------------------------------------------------------------------------*/
    _isModel: function (entity) {
      return entity instanceof Backbone.Model;
    },
    _getUUID: function (obj, shouldGenerate) {
      obj = obj || {};
      var uuid = obj.__uuid || this.__uuid;
      delete obj.__uuid;
      return !uuid && shouldGenerate ? this._generateUUID() : uuid;
    },
    _generateUUID: function () {
      var d = new Date().getTime();
      var uuid = 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
        var r = (d + Math.random() * 16) % 16 | 0;
        d = Math.floor(d / 16);
        return (c === 'x' ? r : r & 3 | 8).toString(16);
      });
      return uuid;
    },
    _getCacheKey: function (entity, name) {
      if (!this._isModel(entity)) {
        return '__' + name;
      }
      return name ? entity.__uuid + '-__' + name : entity.__uuid;
    }
  };
}({});

return backbonelocalcache;


}));