/*!
 * test/backbone.localcache.js
 */


define(function (require) {


/* -----------------------------------------------------------------------------
 * dependencies
 * ---------------------------------------------------------------------------*/

var assert = require('proclaim');
var sinon = require('sinon');
var Backbone = require('backbone');
var mixins = require('mixins/mixins');
var backboneParent = require('backbone.parent/backbone.parent');
var backboneLocalcache = require('backbone.localcache');


/* -----------------------------------------------------------------------------
 * reusable
 * ---------------------------------------------------------------------------*/

var createEntityClass = function (EntityClass, attachMeta, cacheName) {
  return mixins.mix(EntityClass.extend({
    cacheName: _.isUndefined(cacheName) ? 'test' : null,
    constructor: function (attrs, options) {
      if (attachMeta) {
        this.attachChildInstance(new Backbone.Model(this.__attachments.meta), 'meta');
        delete options.meta;
      }

      EntityClass.prototype.constructor.apply(this, arguments);
    }
  }), backboneParent, backboneLocalcache);
};

var CachedNoMetaModel = createEntityClass(Backbone.Model);
var CachedNoNameModel = createEntityClass(Backbone.Model, true, null);
var CachedModel = createEntityClass(Backbone.Model, true);
var CachedCollection = createEntityClass(Backbone.Collection, true);
var CachedChildCollection = CachedCollection.extend({ model: CachedNoNameModel });

var attrs1 = { 'id': 1, 'prop': 'val1' };
var attrs2 = { 'id': 2, 'prop': 'val2' };
var meta1 = { 'id': 1, 'prop': 'meta1' };
var meta2 = { 'id': 2, 'prop': 'meta2' };
var changedAttrs1 = { 'id': 1, 'prop': 'changed-val1' };
var changedAttrs2 = { 'id': 2, 'prop': 'changed-val2' };
var changedMeta1 = { 'id': 1, 'prop': 'changed-meta1' };
var changedMeta2 = { 'id': 2, 'prop': 'changed-meta2' };

var getKey = function (entity, name) {
  var cacheKeys = ['lscache-test'];
  var entityKey = _.isString(entity) ? entity : entity.__uuid;

  if (entityKey) { cacheKeys.push(entityKey); }
  if (name) { cacheKeys.push('__' + name); }

  return cacheKeys.join('-');
};

var assertMetaSynced = function (meta, cache, attrs, key) {
  assert.deepEqual(meta.attributes, attrs);
  assert.deepEqual(meta.attributes, cache.get(key.replace('lscache-test-', '')));
  assert.deepEqual(meta.attributes, JSON.parse(localStorage[key]));
};

var assertModelSynced = function (model, cache, attrs, key) {
  assert.deepEqual(model.attributes, attrs);
  assert.deepEqual(model.attributes, cache.get(model.__uuid));
  assert.deepEqual(model.attributes, JSON.parse(localStorage[key]));
};

var assertCollectionSynced = function (collection, definitions) {
  assert.equal(definitions.length, collection.models.length);

  _.each(definitions, function (definition) {
    assertModelSynced.apply(this, definition);
  });
};


/* -----------------------------------------------------------------------------
 * test
 * ---------------------------------------------------------------------------*/

describe('backbone.localcache', function () {

  beforeEach(function () {
    localStorage.clear();
  });

  describe('cache configuration', function () {

    beforeEach(function () {
      this.defaultsSpy = sinon.spy(CachedModel.prototype, 'getCacheDefaults');
      this.bucketSpy = sinon.spy(CachedModel.prototype, 'getBucketName');
      this.options = {};

      this.model = new CachedModel(attrs1, this.options);
    });

    afterEach(function () {
      this.defaultsSpy.restore();
      this.bucketSpy.restore();
    });

    it('Should pass attrs and options to getCacheDefaults.', function () {
      assert.isTrue(this.defaultsSpy.calledWith(attrs1, this.options));
    });

    it('Should pass attrs and options to getBucketName.', function () {
      assert.isTrue(this.bucketSpy.calledWith(attrs1, this.options));
    });

  });

  describe('model', function () {

    it('Should save data to cache on instantiation.', function () {
      var model = new CachedModel(attrs1);

      assertModelSynced(model, model.cache, attrs1, getKey(model));
    });

    it('Should save data to cache on set.', function () {
      var model = new CachedModel();
      model.set(attrs1);

      assertModelSynced(model, model.cache, attrs1, getKey(model));
    });

    it('Should save meta to cache.', function () {
      var model = new CachedModel(attrs1);
      model.meta.set(meta1);

      assertMetaSynced(model.meta, model.cache, meta1, getKey(model, 'meta'));
    });

    it('Should update data in cache.', function () {
      var model = new CachedModel(attrs1);
      model.set(changedAttrs1);

      assertModelSynced(model, model.cache, changedAttrs1, getKey(model));
    });

    it('Should remove data in cache.', function () {
      var model = new CachedModel(attrs1);
      model.clear();

      assert.isNull(model.cache.get(model.__uuid));
    });

    it('Should load data from cache.', function () {
      localStorage[getKey(attrs1)] = JSON.stringify(attrs1);
      var model = new CachedModel(attrs1);

      assertModelSynced(model, model.cache, attrs1, getKey(model));
    });

    it('Should load meta from cache.', function () {
      localStorage[getKey('1', 'meta')] = JSON.stringify(meta1);
      var model = new CachedModel(attrs1, { __uuid: '1' });

      assertMetaSynced(model.meta, model.cache, meta1, getKey(model, 'meta'));
    });

    it('Should mixin meta from cache.', function () {
      localStorage[getKey('1', 'meta')] = JSON.stringify(meta1);
      var additiveMeta = { prop: 'meta-overwrite', prop2: 'meta2' };
      var model = new CachedModel(attrs1, { meta: additiveMeta, __uuid: '1' });
      var expected = _.extend(meta1, additiveMeta);

      assertMetaSynced(model.meta, model.cache, expected, getKey(model, 'meta'));
    });

    it('Should save meta to cache on instantiation.', function () {
      var model = new CachedModel(attrs1, { meta: meta1 });

      assertMetaSynced(model.meta, model.cache, meta1, getKey(model, 'meta'));
    });

    it('Should filter meta out while loading meta attributes.', function () {
      localStorage[getKey(attrs1, 'meta')] = JSON.stringify(meta1);
      localStorage[getKey(attrs1)] = JSON.stringify(attrs1);
      var model = new CachedModel(attrs1);

      assertModelSynced(model, model.cache, attrs1, getKey(model));
    });

    it('Should be capable of instantiating even if attachment does not exit.', function () {
      var model = new CachedNoMetaModel(attrs1);

      assertModelSynced(model, model.cache, attrs1, getKey(model));
    });

  });

  describe('collection', function () {

    it('Should save models to cache on instantiation.', function () {
      var collection = new CachedCollection([attrs1, attrs2]);

      assertCollectionSynced(collection, [
        [collection.models[0], collection.cache, attrs1, getKey(collection.models[0])],
        [collection.models[1], collection.cache, attrs2, getKey(collection.models[1])]
      ]);
    });

    it('Should save models to cache on add.', function () {
      var collection = new CachedCollection();
      collection.add([attrs1, attrs2]);

      assertCollectionSynced(collection, [
        [collection.models[0], collection.cache, attrs1, getKey(collection.models[0])],
        [collection.models[1], collection.cache, attrs2, getKey(collection.models[1])]
      ]);
    });

    it('Should save meta to cache.', function () {
      var collection = new CachedCollection();
      collection.meta.set(meta1);

      assertMetaSynced(collection.meta, collection.cache, meta1, getKey(collection, 'meta'));
    });

    it('Should update models in cache.', function () {
      var collection = new CachedCollection();
      collection.add([attrs1, attrs2]);
      collection.get(1).set(changedAttrs1);

      assertCollectionSynced(collection, [
        [collection.models[0], collection.cache, changedAttrs1, getKey(collection.models[0])],
        [collection.models[1], collection.cache, attrs2, getKey(collection.models[1])]
      ]);
    });

    it('Should remove models from cache.', function () {
      var collection = new CachedCollection([attrs1, attrs2]);
      collection.remove(attrs2);

      assertCollectionSynced(collection, [
        [collection.models[0], collection.cache, attrs1, getKey(collection.models[0])]
      ]);
    });

    it('Should remove all models from cache.', function () {
      var collection = new CachedCollection([attrs1, attrs2]);
      collection.reset();

      assertCollectionSynced(collection, []);
    });

    it('Should load data from cache.', function () {
      localStorage[getKey('1')] = JSON.stringify(attrs1);
      localStorage[getKey('2')] = JSON.stringify(attrs2);
      var collection = new CachedCollection();

      assertCollectionSynced(collection, [
        [collection.models[0], collection.cache, attrs1, getKey(collection.models[0])],
        [collection.models[1], collection.cache, attrs2, getKey(collection.models[1])]
      ]);
    });

    it('Should load meta from cache.', function () {
      localStorage[getKey(false, 'meta')] = JSON.stringify(meta1);
      var collection = new CachedCollection();

      assertMetaSynced(collection.meta, collection.cache, meta1, getKey(collection, 'meta'));
    });

    it('Should mixin meta from cache.', function () {
      localStorage[getKey(false, 'meta')] = JSON.stringify(meta1);
      var additiveMeta = { prop: 'meta-overwrite', prop2: 'meta2' };
      var collection = new CachedCollection({}, { meta: additiveMeta });
      var expected = _.extend(meta1, additiveMeta);

      assertMetaSynced(collection.meta, collection.cache, expected, getKey(collection, 'meta'));
    });

    it('Should save meta to cache on instantiation.', function () {
      var collection = new CachedCollection({}, { meta: meta1 });

      assertMetaSynced(collection.meta, collection.cache, meta1, getKey(collection, 'meta'));
    });

    it('Should filter meta out while loading meta.', function () {
      localStorage[getKey(false, 'meta')] = JSON.stringify(meta1);
      localStorage[getKey('1')] = JSON.stringify(attrs1);
      localStorage[getKey('2')] = JSON.stringify(attrs2);
      var collection = new CachedCollection();

      assertMetaSynced(collection.meta, collection.cache, meta1, getKey(collection, 'meta'));
      assertCollectionSynced(collection, [
        [collection.models[0], collection.cache, attrs1, getKey(collection.models[0])],
        [collection.models[1], collection.cache, attrs2, getKey(collection.models[1])]
      ]);
    });

  });

  describe('child attachments', function () {

    it('Should load meta for child from cache.', function () {
      localStorage[getKey('1')] = JSON.stringify(attrs1);
      localStorage[getKey('1', 'meta')] = JSON.stringify(meta1);
      localStorage[getKey('2')] = JSON.stringify(attrs2);
      localStorage[getKey('2', 'meta')] = JSON.stringify(meta2);

      var collection = new CachedChildCollection();
      var model1 = collection.models[0];
      var model2 = collection.models[1];

      assertCollectionSynced(collection, [
        [model1, collection.cache, attrs1, getKey(model1)],
        [model2, collection.cache, attrs2, getKey(model2)]
      ]);

      assertMetaSynced(collection.meta, collection.cache, {}, getKey(collection, 'meta'));
      assertMetaSynced(model1.meta, model1.cache, meta1, getKey(model1, 'meta'));
      assertMetaSynced(model2.meta, model2.cache, meta2, getKey(model2, 'meta'));
    });

    it('Should save meta for child to cache.', function () {
      var collection = new CachedChildCollection([attrs1, attrs2]);
      var model1 = collection.models[0];
      var model2 = collection.models[1];

      model1.meta.set(meta1);
      model2.meta.set(meta2);

      assertCollectionSynced(collection, [
        [model1, collection.cache, attrs1, getKey(model1)],
        [model2, collection.cache, attrs2, getKey(model2)]
      ]);

      assertMetaSynced(collection.meta, collection.cache, {}, getKey(collection, 'meta'));
      assertMetaSynced(model1.meta, model1.cache, meta1, getKey(model1, 'meta'));
      assertMetaSynced(model2.meta, model2.cache, meta2, getKey(model2, 'meta'));
    });

  });
});


});